
#define _USE_MATH_DEFINES
#include <map>
#include <float.h>
#include <fstream>

#include "global.h"
#include "matrixd.h"
#include "bary.h"

#include "sparse_row_mat.h"
#include "conjugate_gradient.h"

vect2d perp(const vect2d &a)
{
	return vect2d(-a.v[1], a.v[0]);
}




void calcBaryMLS(double *bary, vect2d &x, vector<vect2d> &pv, vector<vect2i> &p)
{
	double epsilon = 1e-3;

	int n = pv.size();
	if (n < 3)
		return;

	// calculate distances from line and projection onto line
	vector<double> dists(n);
	vector<double> projs(n);
	for (int i = 0; i < n; i++)
	{
		vect2d vv = pv[p[i][1]] - pv[p[i][0]];
		vect2d xx = x - pv[p[i][0]];
		projs[i] = xx*vv/vv.length2();
		dists[i] = (xx - vv*projs[i]).length();
	}

	// fix degenerate cases (force to interpolate the line)
	for (int i = 0; i < n; i++)
	{
		if (dists[i] < epsilon && projs[i] >= 0 && projs[i] <= 1)
		{
			bary[p[i][0]] += 1 - projs[i];
			bary[p[i][1]] += projs[i];
			return;
		}
	}

	// calculate inner weight  matrix
	vector<Matrixd> weights(n);
	for (int i = 0; i < n; i++)
	{
		weights[i].resize(2, 2);

		if (g.weight_function == length_difference) // integral that doesn't depend on points
		{
			double v = !(x - pv[p[i][0]]) + !(x - pv[p[i][1]]) - !(pv[p[i][1]] - pv[p[i][0]]);
			double w;
			if (v <= 0)
				w = 1e30;
			else
				w = 1.0 / (v*v);

			weights[i](0,0) = w/3;
			weights[i](0,1) = w/6;
			weights[i](1,0) = w/6;
			weights[i](1,1) = w/3;
		}
		else if (g.weight_function == reciprocal_distance) // MLS method in paper
		{
			if (dists[i] < epsilon && (projs[i] < 0 || projs[i] > 1))
			{
				vect2d &a = pv[p[i][0]];
				vect2d &b = pv[p[i][1]];
				vect2d ba = b-a;
				double bl2 = ba.length2();
				double bl5 = bl2*bl2*sqrt(bl2);
				vect2d vb = x-b;
				vect2d av = x-a;
				double c = vb*ba;
				double d = av*ba;

				weights[i](0,0) = bl5/(3*c*d*d*d);
				weights[i](0,1) = bl5/(6*c*c*d*d);
				weights[i](1,0) = weights[i](0,1);
				weights[i](1,1) = bl5/(3*c*c*c*d);
			}
			else
			{
				vect2d &a = pv[p[i][0]];
				vect2d &b = pv[p[i][1]];
				double delta = perp(a - x)*(a - b);
				double theta = atan( ( (b-x)*(b-a) ) / ( perp(b-x)*(b-a) ) ) - atan( ( (a-x)*(a-b) ) / ( perp(a-x)*(a-b) ) );
				double B00 = (a-x)*(a-x);
				double B01 = (a-x)*(x-b);
				double B11 = (x-b)*(x-b);
				double pre = !(a-b) / (2*delta*delta);

				weights[i](0,0) = pre * (B01/B00 - B11*theta/delta);
				weights[i](0,1) = pre * (1 - B01*theta/delta);
				weights[i](1,0) = weights[i](0,1);
				weights[i](1,1) = pre * (B01/B11 - B00*theta/delta);
			}
		}
	}

	// calculate constant
	Matrixd M(3, 3), P(3, 2);
	M.set_zero();
	P(2,0) = P(2,1) = 1;
	for (int i = 0; i < n; i++)
	{
		P.set_col(0, pv[p[i][0]].v, 2);
		P.set_col(1, pv[p[i][1]].v, 2);

		M = M + P*weights[i]*P.t();
	}

	Matrixd X(3,1);
	X.set_col(0, x.v, 2);
	X[2] = 1;
	Matrixd k = M.inverse()*X;

	// calculate barycentric coordinates
	P.resize(2,3);
	P(0,2) = P(1,2) = 1;
	for (int i = 0; i < n; i++)
	{
		P.set_row(0, pv[p[i][0]].v, 2);
		P.set_row(1, pv[p[i][1]].v, 2);

		Matrixd alpha = weights[i]*P*k;

		bary[p[i][0]] += alpha[0];
		bary[p[i][1]] += alpha[1];
	}

	// bary coords must sum to one
	double sum = 0;
	for (int i = 0; i < n; i++)
		sum += bary[i];
	for (int i = 0; i < n; i++)
		bary[i] /= sum;
}

void calcBaryPts(double *bary, vect2d &x, vector<vect2d> &p)
{
	int n = p.size();
	if (n < 3)
		return;

	// calculate weights
	vector<double> weights(n, 0);
	for (int i = 0; i < n; i++)
	{
		double l2 = (x - p[i]).length2();
		weights[i] = 1.0 / l2;

		if (l2 < 1e-6)
		{
			for (int j = 0; j < n; j++)
				bary[j] = (j == i ? 1 : 0);
			return;
		}
	}

	// calculate constant
	double a = 0;
	double b = 0;
	double c = 0;
	double d = 0;
	double e = 0;
	double f = 0;
	for (int i = 0; i < n; i++)
	{
		a += weights[i] * p[i][0]*p[i][0];
		b += weights[i] * p[i][1]*p[i][0];
		c += weights[i] * p[i][0];
		d += weights[i] * p[i][1]*p[i][1];
		e += weights[i] * p[i][1];
		f += weights[i];
	}
	double det = -b*b*f + a*d*f - c*c*d + 2*b*c*e - a*e*e;
	vect3f k(-c*d + b*e + d*f*x[0] - e*e*x[0] - b*f*x[1] + c*e*x[1],
		-a*e + c*e*x[0] + b*(c - f*x[0]) - c*c*x[1] + a*f*x[1],
		-b*b + a*d - c*d*x[0] + b*e*x[0] + b*c*x[1] - a*e*x[1]);
	k /= det;

	// calculate barycentric coordinates
	for (int i = 0; i < n; i++)
		bary[i] = weights[i] * (p[i][0]*k[0] + p[i][1]*k[1] + k[2]);

	// sanity check
	double sum = 0;
	for (int i = 0; i < n; i++)
		sum += bary[i];
	//assert(fabs(1 - sum) < 1e-3);
}

void calcBaryDeriv(double *bary, double *bary2, vect2d &x, vector<vect2d> &pv, vector<vect2i> &p, vector<vect2d> &r)
{
	double epsilon = 1e-3;

	int n = pv.size();
	if (n < 3)
		return;

	// calculate distances from line and projection onto line
	vector<double> dists(n);
	vector<double> projs(n);
	for (int i = 0; i < n; i++)
	{
		vect2d vv = pv[p[i][1]] - pv[p[i][0]];
		vect2d xx = x - pv[p[i][0]];
		projs[i] = xx*vv/vv.length2();
		dists[i] = (xx - vv*projs[i]).length();
	}

	// fix degenerate cases (force to interpolate the line)
	for (int i = 0; i < n; i++)
	{
		if (dists[i] < epsilon && projs[i] >= 0 && projs[i] <= 1)
		{
			bary[p[i][0]] += 1 - projs[i];
			bary[p[i][1]] += projs[i];
			return;
		}
	}

	// calculate inner weight  matrix
	vector<Matrixd> weights(n);
	for (int i = 0; i < n; i++)
	{
		weights[i].resize(2, 2);

		if (dists[i] < epsilon && (projs[i] < 0 || projs[i] > 1))
		{
			vect2d &a = pv[p[i][0]];
			vect2d &b = pv[p[i][1]];
			vect2d ba = b-a;
			double bl2 = ba.length2();
			double bl5 = bl2*bl2*sqrt(bl2);
			vect2d vb = x-b;
			vect2d av = x-a;
			double c = vb*ba;
			double d = av*ba;

			weights[i](0,0) = bl5/(3*c*d*d*d);
			weights[i](0,1) = bl5/(6*c*c*d*d);
			weights[i](1,0) = weights[i](0,1);
			weights[i](1,1) = bl5/(3*c*c*c*d);
		}
		else
		{
			vect2d &a = pv[p[i][0]];
			vect2d &b = pv[p[i][1]];
			double delta = perp(a - x)*(a - b);
			double theta = atan( ( (b-x)*(b-a) ) / ( perp(b-x)*(b-a) ) ) - atan( ( (a-x)*(a-b) ) / ( perp(a-x)*(a-b) ) );
			double B00 = (a-x)*(a-x);
			double B01 = (a-x)*(x-b);
			double B11 = (x-b)*(x-b);
			double pre = !(a-b) / (2*delta*delta);

			weights[i](0,0) = pre * (B01/B00 - B11*theta/delta);
			weights[i](0,1) = pre * (1 - B01*theta/delta);
			weights[i](1,0) = weights[i](0,1);
			weights[i](1,1) = pre * (B01/B11 - B00*theta/delta);
		}
	}

	// calculate constant
	Matrixd M(3, 3), P(3, 2), V(3, 2);
	M.set_zero();
	P(2,0) = P(2,1) = 1;
	V(2,0) = V(2,1) = 0;
	for (int i = 0; i < n; i++)
	{
		P.set_col(0, pv[p[i][0]].v, 2);
		P.set_col(1, pv[p[i][1]].v, 2);
		V.set_col(0, r[i].v, 2);
		V.set_col(1, r[i].v, 2);

		M = M + P*weights[i]*P.t() + V*weights[i]*V.t();
	}

	Matrixd X(3,1);
	X.set_col(0, x.v, 2);
	X[2] = 1;
	Matrixd k = M.inverse()*X;

	// calculate barycentric coordinates
	P.resize(2,3);
	P(0,2) = P(1,2) = 1;
	V.resize(2,3);
	V(0,2) = V(1,2) = 0;
	for (int i = 0; i < n; i++)
	{
		P.set_row(0, pv[p[i][0]].v, 2);
		P.set_row(1, pv[p[i][1]].v, 2);
		V.set_row(0, r[i].v, 2);
		V.set_row(1, r[i].v, 2);

		Matrixd alpha = weights[i]*P*k;
		Matrixd beta = weights[i]*V*k;

		bary[p[i][0]] += alpha[0];
		bary[p[i][1]] += alpha[1];

		bary2[i] += beta[0];
		bary2[i] += beta[1];
	}

	// bary coords must sum to one
	double sum = 0;
	for (int i = 0; i < n; i++)
		sum += bary[i];
	for (int i = 0; i < n; i++)
		bary[i] /= sum;
}

void calcBaryWachspress(double *bary, vect2d &x, vector<vect2d> &pv, vector<vect2i> &p)
{
#define vec3(vv) vect3d((vv)[0],(vv)[1],0)
	int n = p.size();
	if (n < 3)
		return;

	for (int i = 0; i < n; i++)
	{
		double a;
		a = (vec3(pv[(i+1)%n] - pv[i]) % vec3(pv[(i+n-1)%n] - pv[i]))[2];
		bary[i] = a;

		for (int jj = 1; jj < n-1; jj++)
		{
			int j = (i + jj) % n;
			int jp = (i + jj + 1) % n;
			a = (vec3(pv[j] - x) % vec3(pv[jp] - x))[2];
			bary[i] *= a;
		}
	}

	// normalize
	double sum = 0;
	for (int i = 0; i < n; i++)
		sum += bary[i];
	for (int i = 0; i < n; i++)
		bary[i] /= sum;
#undef vec3
}

void calcBaryMeanValue(double *bary, vect2d &x, vector<vect2d> &pv, vector<vect2i> &p)
{
	int n = p.size();
	if (n < 3)
		return;

	for (int i = 0; i < n; i++)
		bary[i] = 0;

	for (int i = 0; i < n; i++)
	{
		vect2d v1 = pv[i] - x;
		vect2d v2 = pv[(i+1)%n] - x;

		double d1 = v1.length();
		double d2 = v2.length();

		v1 /= d1;
		v2 /= d2;

		double c = v1*v2;
		double s = v1[0]*v2[1] - v1[1]*v2[0]; // det(v1, v2)
		bary[i] += (1-c)/(s*d1);
		bary[(i+1)%n] += (1-c)/(s*d2);
	}

	// normalize
	double sum = 0;
	for (int i = 0; i < n; i++)
		sum += bary[i];
	for (int i = 0; i < n; i++)
		bary[i] /= sum;
}

void calcBaryHarmonic()
{
	int n_pts = g.tri_pt_pos.size();
	int n_bary = g.poly[0].size();
	int n_tri = g.tri_tris.size();
	int n_inside = 0;

	// build maps
	map<int, int> big2small, small2big;
	for (int i = 0; i < n_pts; i++)
	{
		if (g.tri_pt_border[i])
			continue;

		big2small[i] = n_inside;
		small2big[n_inside] = i;

		n_inside++;
	}
	
	// build matrices
#if 0
	Matrixd ATA(n_inside, n_inside);
	ATA.set_zero();
#else
	SparseRowMat ATA(n_inside, n_inside);
#endif

	Matrixd ATB(n_inside, n_bary);
	ATB.set_zero();

	// set laplacian
	for (int i = 0; i < n_tri; i++)
	{
		for (int j = 0; j < 3; j++)
		{
			int ia = g.tri_tris[i][j];
			int ib = g.tri_tris[i][(j+1)%3];
			int ic = g.tri_tris[i][(j+2)%3];

			if (g.tri_pt_border[ic])
				continue;

			vect2d &pa = g.tri_pt_pos[ia];
			vect2d &pb = g.tri_pt_pos[ib];
			vect2d &pc = g.tri_pt_pos[ic];

			vect2d e = pb - pa;
			vect2d vn(e[1], -e[0]);
			Matrixd N(2,1);
			N(0,0) = vn[0];
			N(1,0) = vn[1];

			vect2d va = pa - pc;
			vect2d vb = pb - pc;
			Matrixd V(2,2);
			V(0,0) = va[0];
			V(1,0) = va[1];
			V(0,1) = vb[0];
			V(1,1) = vb[1];

			double detV = V(0,0)*V(1,1) - V(0,1)*V(1,0);
			Matrixd VI(2,2);
			VI(0,0) = V(1,1) / detV;
			VI(1,0) = -V(1,0) / detV;
			VI(0,1) = -V(0,1) / detV;
			VI(1,1) = V(0,0) / detV;

			Matrixd W = VI * N;
			double wa = W(0,0);
			double wb = W(1,0);

			// add weights
			int inc = big2small[ic];
#if 0
			ATA(inc, inc) -= wa + wb;
#else
			ATA.set(inc, inc, ATA.get(inc, inc) - wa - wb);
#endif

			if (g.tri_pt_border[ia])
			{
				for (int k = 0; k < n_bary; k++)
					ATB(inc, k) -= wa * g.tri_pt_barys[ia*n_bary + k];
			}
			else
			{
				int ina = big2small[ia];
#if 0
				ATA(inc, ina) += wa;
#else
				ATA.set(inc, ina, ATA.get(inc, ina) + wa);
#endif
			}
			
			if (g.tri_pt_border[ib])
			{
				for (int k = 0; k < n_bary; k++)
					ATB(inc, k) -= wb * g.tri_pt_barys[ib*n_bary + k];
			}
			else
			{
				int inb = big2small[ib];
#if 0
				ATA(inc, inb) += wb;
#else
				ATA.set(inc, inb, ATA.get(inc, inb) + wb);
#endif
			}
		}
	}

#if 0
	// solve system to get barys in poly
	Matrixd X = ATA.inverse() * ATB;
	
	// copy solution back into bary coords
	for (int i = 0; i < n_inside; i++)
	{
		int ii = small2big[i];
		if (g.tri_pt_border[ii])
			continue;
		
		for (int k = 0; k < n_bary; k++)
			g.tri_pt_barys[ii*n_bary + k] = X(i, k);
	}
#else
	for (int k = 0; k < n_bary; k++)
	{
		// fill initial guess for X
		DenseVect X(n_inside);
		for (int i = 0; i < n_inside; i++)
		{
			int ii = small2big[i];
			if (g.tri_pt_border[ii])
				continue;
			
			double v = g.tri_pt_barys[ii*n_bary + k];
			if (v < 0)
				v = 0;
			else if (v > 1)
				v = 1;

			X.set(i, v);
		}

		// fill B
		DenseVect B(n_inside);
		for (int i = 0; i < n_inside; i++)
			B.set(i, ATB(i, k));

		// solve for X
		double tol = 1e-6;
		int iters = n_inside;
		conjugate_gradient(ATA, X, B, tol, iters);

		printf("iters = %d/%d, error = %f\n", iters, n_inside, tol);

		// copy solution back into bary coords
		for (int i = 0; i < n_inside; i++)
		{
			int ii = small2big[i];
			if (g.tri_pt_border[ii])
				continue;
			
			double val = X.get(i);
			if (val < 0) // may be negative from numerical imprecision
				val = 0;

			g.tri_pt_barys[ii*n_bary + k] = val;
		}
	}
#endif
}

double calcBaryMaxEntropy_F(vect2d &lambda, vector<double> &m, vector<vect2d> &vrel)
{
	double z = 0;
	for (int i = 0; i < m.size(); i++)
		z += m[i] * exp(-lambda * vrel[i]);
	return z;
}

void calcBaryMaxEntropy(double *bary, vect2d &x, vector<vect2d> &pv, vector<vect2i> &p)
{
	int n = p.size();
	if (n < 3)
		return;

	double epsilon = 1e-6;
	vect2d lambda = 0;

	// precalculate some stuff
	vector<vect2d> vrel(n);
	vector<double> rho(n);
	vector<double> pi(n);
	vector<double> m(n);

	for (int i = 0; i < n; i++)
	{
		int ip = (i+1)%n;

		vrel[i] = pv[i] - x;
		rho[i] = (pv[i] - x).length() + (pv[ip] - x).length() - (pv[i] - pv[ip]).length();
	}

	double pi_sum = 0;
	for (int i = 0; i < n; i++)
	{
		int im = (i+n-1)%n;
		pi[i] = 1.0 / (rho[im]*rho[i]);
		pi_sum += pi[i];
	}
	
	for (int i = 0; i < n; i++)
	{
		m[i] = pi[i] / pi_sum;
	}

	// perform minimization
	//printf("before = %f,%f -> %f\n", lambda[0], lambda[1], calcBaryMaxEntropy_F(lambda, m, vrel));
	//for (int i = 0; i < 5; i++)
	//{
	//	for (int j = 0; j < 5; j++)
	//	{
	//		vect2d test(i,j);
	//		printf("\t%f,%f -> %f\n", test[0], test[1], calcBaryMaxEntropy_F(test, m, vrel));
	//	}
	//}

	for (int i = 0; i < 20; i++)
	{
		// calc the function value, gradient, and hessian
		vect2d dx(epsilon, 0);
		vect2d dy(0, epsilon);
		double f = calcBaryMaxEntropy_F(lambda, m, vrel);
		vect2d g;
		g[0] = (calcBaryMaxEntropy_F(lambda + dx, m, vrel) - calcBaryMaxEntropy_F(lambda - dx, m, vrel)) / (2 * epsilon);
		g[1] = (calcBaryMaxEntropy_F(lambda + dy, m, vrel) - calcBaryMaxEntropy_F(lambda - dy, m, vrel)) / (2 * epsilon);
		vect<2, vect2d> h;
		h[0][0] = (calcBaryMaxEntropy_F(lambda + dx, m, vrel) - 2*f + calcBaryMaxEntropy_F(lambda - dx, m, vrel)) / (epsilon * epsilon);
		h[1][1] = (calcBaryMaxEntropy_F(lambda + dy, m, vrel) - 2*f + calcBaryMaxEntropy_F(lambda - dy, m, vrel)) / (epsilon * epsilon);
		h[0][1] = h[1][0] = ((calcBaryMaxEntropy_F(lambda + dy + dx, m, vrel) - calcBaryMaxEntropy_F(lambda - dy + dx, m, vrel)) - (calcBaryMaxEntropy_F(lambda + dy - dx, m, vrel) - calcBaryMaxEntropy_F(lambda - dy - dx, m, vrel))) / (4 * epsilon * epsilon);

		// choose search direction
		double deth = h[0][0]*h[1][1] - h[0][1]*h[1][0];
		if (deth < epsilon)
			deth = epsilon;

		vect<2, vect2d> h_inv;
		h_inv[0][0] = h[1][1] / deth;
		h_inv[1][1] = h[0][0] / deth;
		h_inv[0][1] = -h[0][1] / deth;
		h_inv[1][0] = -h[1][0] / deth;

		vect2d step(-h_inv[0] * g, -h_inv[1] * g); // step is a newton step and should try to go to the min
		double step_len2 = step.length2();

		if (!_finite(step[0]) || !_finite(step[1]))
		{
			printf("bad step taken, aborting.\n");
			printf("g = %f, %f\n", g[0], g[1]);
			printf("deth = %f\n", deth);
			break;
		}

		// update
		if (step_len2 > .01)
		{
			//**** step distance is too large, so perform a line search

			// first try to bracket the value
			vect2d pts[3];
			double vals[3];

			pts[0] = lambda;
			vals[0] = calcBaryMaxEntropy_F(pts[0], m, vrel);
			pts[1] = lambda + step;
			vals[1] = calcBaryMaxEntropy_F(pts[1], m, vrel);

			if (vals[1] > vals[0])
			{
				swap(vals[0], vals[1]);
				swap(pts[0], pts[1]);
				step = -step;
			}

			do
			{
				pts[2] = pts[1] + step;
				vals[2] = calcBaryMaxEntropy_F(pts[2], m, vrel);
				step *= 2;
			} while (vals[2] < vals[1]);

			// perform golden search for minimum along line
			for (int k = 0; k < 100; k++)
			{
				// split the larger interval
				vect2d a = pts[1] - pts[0];
				vect2d b = pts[2] - pts[1];
				double a2 = a*a;
				double b2 = b*b;

				if (a2 + b2 < 1e-6)
					break;

				if (a2 > b2)
				{
					vect2d c = pts[0] + a * .38197;
					double vc = calcBaryMaxEntropy_F(c, m, vrel);
					if (vc < vals[1])
					{
						pts[2] = pts[1];
						vals[2] = vals[1];
						pts[1] = c;
						vals[1] = vc;
					}
					else
					{
						pts[0] = c;
						vals[0] = vc;
					}
				}
				else
				{
					vect2d c = pts[2] - b * .38197;
					double vc = calcBaryMaxEntropy_F(c, m, vrel);
					if (vc < vals[1])
					{
						pts[0] = pts[1];
						vals[0] = vals[1];
						pts[1] = c;
						vals[1] = vc;
					}
					else
					{
						pts[2] = c;
						vals[2] = vc;
					}
				}
			}

			// set lambda to the minimum that was found
			lambda = pts[1];
		}
		else
			lambda += .95*step; // perform damped newton step

		if (g.length() < epsilon)
			break;
	}
	//printf("after = %f,%f -> %f\n", lambda[0], lambda[1], calcBaryMaxEntropy_F(lambda, m, vrel));

	// convert from lambda to b
	double z = 0;
	for (int i = 0; i < n; i++)
	{
		bary[i] = m[i] * exp(-lambda * vrel[i]);
		z += bary[i];
	}
	for (int i = 0; i < n; i++)
		bary[i] /= z;

	//// check linear precision
	//vect2d pt = 0;
	//for (int i = 0; i < n; i++)
	//	pt += bary[i] * pv[i];
	//printf("%f, %f -> %f, %f\n", x[0], x[1], pt[0], pt[1]);
}

void calcBaryFile()
{
	int n = g.poly[0].size();

	// read coordinates from a file
	char fn[1024];
	sprintf(fn, "spiral_m3_a%d.txt", g.alpha);
	ifstream f(fn);
	if (f.fail())
	{
		printf("FAILED to open file %s\n", fn);
		return;
	}

	int bc_num = 1;
	for (int i = 0; i < g.tri_pt_pos.size(); i++)
	{
		if (!g.tri_pt_border[i])
		{
			vect2d pos, actual = g.tri_pt_pos[i];
			f >> pos[0] >> pos[1];
			
			assert(!(pos - actual) < .01);

			for (int j = 0; j < n; j++)
				f >> g.tri_pt_barys[i*n + j];

			bc_num++;
		}
		else
		{
			int asdffsda = 0;
		}
	}
	f.close();
}

void calcBary(double *bary, double *bary2, vect2d &x, vector<vect2d> &pv, vector<vect2i> &p, vector<vect2d> &r)
{
	if (g.method == points_linear)
		calcBaryPts(bary, x, pv);
	else if (g.method == lines_linear_old)
	{
		if (g.deriv == 0)
			calcBaryMLS(bary, x, pv, p);
		else
			calcBaryDeriv(bary, bary2, x, pv, p, r);
	}
	else if (g.method == lines_linear)
	{
		if (g.alpha == 1)
			calcBary_Alpha1(bary, bary2, x, pv, p, r);
		else if (g.alpha == 2)
			calcBary_Alpha2(bary, bary2, x, pv, p, r);
		else if (g.alpha == 3)
			calcBary_Alpha3(bary, bary2, x, pv, p, r);
	}
	else if (g.method == lines_quad)
	{
		if (g.alpha == 1)
			calcBaryQuad_Alpha1(bary, bary2, x, pv, p, r);
		else if (g.alpha == 2)
			calcBaryQuad_Alpha2(bary, bary2, x, pv, p, r);
	}
	else if (g.method == wachspress)
		calcBaryWachspress(bary, x, pv, p);
	else if (g.method == mean_value)
		calcBaryMeanValue(bary, x, pv, p);
	else if (g.method == maximum_entropy)
		calcBaryMaxEntropy(bary, x, pv, p);
}

void calcBaryCoords()
{
	int n = g.poly[0].size();

	// construct line segment indices
	vector<vect2i> lines(n);
	for (int i = 0; i < n; i++)
	{
		int ip = (i+1)%n;

		lines[i][0] = i;
		lines[i][1] = ip;
	}

	//=====================================================
	// Regular grid
	//=====================================================

	//// calculate the coordinates
	//g.allocBarycoords();
	//for (int x = 0; x <= g.N; x++)
	//{
	//	for (int y = 0; y <= g.N; y++)
	//	{
	//		vect2d p((double)x/g.N, (double)y/g.N);

	//		double *bary = &g.barycoords[((y*(g.N+1))+x)*n];
	//		double *bary2 = &g.barycoords2[((y*(g.N+1))+x)*n];

	//		calcBary(bary, bary2, p, g.poly[0], lines, g.norms[0]);
	//	}
	//}
	
	//=====================================================
	// Trianglulated polygon
	//=====================================================

	if (g.method == from_file)
		calcBaryFile();
	else if (g.method == points_linear)
	{
		for (int i = 0; i < g.tri_pt_pos.size(); i++)
		{
			double *bary = &g.tri_pt_barys[i * n];
			for (int j = 0; j < n; j++)
				bary[j] = 0;

			calcBary(bary, 0, g.tri_pt_pos[i], g.poly[0], lines, g.norms[0]);
		}
	}
	else if (g.method == harmonic)
	{
		// calc with another mothod as an initial guess
		for (int i = 0; i < g.tri_pt_pos.size(); i++)
		{
			if (g.tri_pt_border[i])
				continue;

			double *bary = &g.tri_pt_barys[i * n];
			for (int j = 0; j < n; j++)
				bary[j] = 0;

			calcBaryMeanValue(bary, g.tri_pt_pos[i], g.poly[0], lines);
			//calcBary_Alpha2_Plin_Qlin(bary, 0, g.tri_pt_pos[i], g.poly[0], lines, g.norms[0]);
		}

		// calc harmonic coordinates
		calcBaryHarmonic();
	}
	else
	{
		for (int i = 0; i < g.tri_pt_pos.size(); i++)
		{
			if (g.tri_pt_border[i])
				continue;
#ifdef QUADRIC_BOUNDARY_DERIV_FIGURE
			double *bary = &g.tri_pt_barys[i * n*3];
			for (int j = 0; j < n*3; j++)
				bary[j] = 0;
#else
			double *bary = &g.tri_pt_barys[i * n];
			for (int j = 0; j < n; j++)
				bary[j] = 0;
#endif
			double *bary2 = &g.tri_pt_barys2[i * n*3]; //needs 3x coords per line for quad precision
			for (int j = 0; j < n*3; j++)
				bary2[j] = 0;

			calcBary(bary, bary2, g.tri_pt_pos[i], g.poly[0], lines, g.norms[0]);
			
			//if (g.deriv == 1)
			//{
			//	// check that the bary coords sum to one
			//	double sum = 0;
			//	for (int i = 0; i < n; i++)
			//		sum += bary[i];
			//	printf("barysum = %f\n", sum);
			//		
			//	// check that the DERIVATIVE bary coords sum to zero
			//	sum = 0;
			//	for (int i = 0; i < 3*n; i++)
			//		sum += bary2[i];
			//	printf("bary2sum = %f\n", sum);
			//}
		}
	}
	
#ifdef QUADRIC_BOUNDARY_DERIV_FIGURE
	// set border values
	for (int i = 0; i < g.tri_pt_pos.size(); i++)
	{
		if (!g.tri_pt_border[i])
			continue;

		double *bary = &g.tri_pt_barys[i * n*3];
		for (int j = 0; j < n*3; j++)
			bary[j] = 0;

		vect2d &pt = g.tri_pt_pos[i];

		// check if it is a line
		for (int j = 0; j < n; j++)
		{
			vect2d p0 = g.poly[0][j], p1 = g.poly[0][(j + 1) % n];

			double d0 = (pt - p0).length();
			double d1 = (pt - p1).length();
			double dt = (p1 - p0).length();
			if ( d0 + d1 - dt < 1e-8)
			{
				float t = d0 / dt;
				bary[j*3 + 0] = (1-t)*(1-t);
				bary[j*3 + 1] = 2*(1-t)*t;
				bary[j*3 + 2] = t*t;
			}
		}

		// check if it is a vertex
		for (int j = 0; j < n; j++)
		{
			if ( (pt - g.poly[0][j]).length() < 1e-4)
			{
				bary[j*3] = .5;
				bary[(j*3 - 1 + n*3) % (n*3)] = .5;
			}
		}

	}
#endif

	//// test derivatives
	//if (g.deriv == 1)
	//{
	//	double *bary = new double [n];
	//	double *bary2 = new double [n*3];
	//	for (int j = 0; j < n*3; j++)
	//	{
	//	for (int i = 0; i < n; i++)
	//	{
	//		double eps = 5e-3;
	//		int ip = (i+1) % n;
	//		vect2d c = (g.poly[0][i] + g.poly[0][ip]) * .5;

	//		memset(bary, 0, sizeof(double)*n);
	//		memset(bary2, 0, sizeof(double)*n*3);
	//		calcBary(bary, bary2, c, g.poly[0], lines, g.norms[0]);
	//		double v1 = bary2[j];
	//		
	//		c += g.norms[0][i] * eps;
	//		memset(bary, 0, sizeof(double)*n);
	//		memset(bary2, 0, sizeof(double)*n*3);
	//		calcBary(bary, bary2, c, g.poly[0], lines, g.norms[0]);
	//		double v2 = bary2[j];

	//		double deriv = (v2 - v1) / eps;
	//		printf("deriv %d = %f\n", i, deriv);
	//	}
	//	}
	//	delete[] bary;
	//	delete[] bary2;
	//}
	//else if (g.deriv == 2)
	//{
	//	double *bary = new double [n];
	//	double *bary2 = new double [n*3];
	//	for (int j = 0; j < n*3; j++)
	//	{
	//	for (int i = 0; i < n; i++)
	//	{
	//		double eps = 5e-3;
	//		int ip = (i+1) % n;
	//		vect2d c = (g.poly[0][i] + g.poly[0][ip]) * .5;

	//		memset(bary, 0, sizeof(double)*n);
	//		memset(bary2, 0, sizeof(double)*n*3);
	//		calcBary(bary, bary2, c, g.poly[0], lines, g.norms[0]);
	//		double v1 = bary2[j];
	//		
	//		c += g.norms[0][i] * eps;
	//		memset(bary, 0, sizeof(double)*n);
	//		memset(bary2, 0, sizeof(double)*n*3);
	//		calcBary(bary, bary2, c, g.poly[0], lines, g.norms[0]);
	//		double v2 = bary2[j];

	//		c += g.norms[0][i] * 2*eps;
	//		memset(bary, 0, sizeof(double)*n);
	//		memset(bary2, 0, sizeof(double)*n*3);
	//		calcBary(bary, bary2, c, g.poly[0], lines, g.norms[0]);
	//		double v3 = bary2[j];

	//		double deriv = (v3 - 2*v2 + v1) / (eps*eps);
	//		printf("deriv %d = %f\n", i, deriv);
	//	}
	//	}
	//	delete[] bary;
	//	delete[] bary2;
	//}
}
