// Author: Josiah Manson

#include <cstdio>
#include <vector>
#include <algorithm>
#include <map>
#include <list>

#define _USE_MATH_DEFINES
#include <cmath>

#include "vect.h"
#include "global.h"
#include "opengl.h"
#include "bary.h"

#include "triangle.h"

using namespace std;

extern Global g;

#include "SDL.h"

const double pi = M_PI;

int contour_steps = 10;
int contour_zero = contour_steps / 5;
int contour_vals = contour_steps + 2*contour_zero + 1;

using namespace std;

void calcTriangulation()
{
	int n = g.poly[0].size();

	// create input to triangulation algorithm
	triangulateio in, out;
	memset(&in, 0, sizeof(triangulateio));
	memset(&out, 0, sizeof(triangulateio));

	in.numberofpoints = n;
	in.numberofsegments = n;
	in.pointlist = (double*)malloc(n * 2 * sizeof(double));
	in.pointmarkerlist = (int*)malloc(n * sizeof(int));
	in.segmentlist = (int*)malloc(n * 2 * sizeof(int));
	in.segmentmarkerlist = (int*)malloc(n * sizeof(int));

	for (int i = 0; i < n; i++)
	{
		in.pointlist[i*2 + 0] = g.poly[0][i][0];
		in.pointlist[i*2 + 1] = g.poly[0][i][1];

		in.pointmarkerlist[i] = 1;

		in.segmentlist[i*2 + 0] = i;
		in.segmentlist[i*2 + 1] = (i+1) % n;
		
		in.segmentmarkerlist[i] = 1;
	}

	// calculate the triangulation
	triangulate("QzpDq0ea.0025", &in, &out, 0);

	// store the results
	g.tri_edges.clear();
	g.tri_tris.clear();
	g.tri_pt_pos.clear();
	g.tri_pt_barys.clear();
	g.tri_pt_barys2.clear();
	g.tri_pt_border.clear();

	for (int i = 0; i < out.numberofpoints; i++)
		g.tri_pt_pos.push_back(vect2d(out.pointlist[i*2 + 0], out.pointlist[i*2 + 1]));
	
	for (int i = 0; i < out.numberofedges; i++)
		g.tri_edges.push_back(vect2i(out.edgelist[i*2 + 0], out.edgelist[i*2 + 1]));
	
	for (int i = 0; i < out.numberoftriangles; i++)
		g.tri_tris.push_back(vect3i(out.trianglelist[i*3 + 0], out.trianglelist[i*3 + 1], out.trianglelist[i*3 + 2]));

	// calculate the interpolated values along the boundary
	for (int i = 0; i < n*out.numberofpoints; i++)
	{
		g.tri_pt_border.push_back(false);
		
		g.tri_pt_barys.push_back(0);
#ifdef QUADRIC_BOUNDARY_DERIV_FIGURE
		g.tri_pt_barys.push_back(0);
		g.tri_pt_barys.push_back(0);
#endif

		g.tri_pt_barys2.push_back(0);
		g.tri_pt_barys2.push_back(0);
		g.tri_pt_barys2.push_back(0);
	}

	for (int i = 0; i < out.numberofsegments; i++)
	{
		for (int jj = 0; jj < 2; jj++)
		{
			int j = out.segmentlist[i*2 + jj];
			g.tri_pt_border[j] = true;
			vect2d &pt = g.tri_pt_pos[j];

			// find the line that the point is on
			for (int k = 0; k < n; k++)
			{
				int kp = (k+1)%n;
				vect2d &p0 = g.poly[0][k];
				vect2d &p1 = g.poly[0][kp];

				double d0 = (p0 - pt).length();
				double d1 = (p1 - pt).length();
				double d = (p1 - p0).length();

				if (d0 + d1 < 1e-6 + d)
				{
					// set the bary coords
					for (int c = 0; c < n; c++)
					{
						double v = (c==k?1:0)*d1/d + (c==kp?1:0)*d0/d;
						g.tri_pt_barys[j*n + c] = v;
					}
				}
			}
		}
	}

	//// write out a file with the point locations (for mathematica to evaluate)
	//FILE *f = fopen("point_locations.txt", "w");

	//fprintf(f, "{");
	//for (int i = 0; i < g.tri_pt_pos.size(); i++)
	//{
	//	if (!g.tri_pt_border[i])
	//		fprintf(f, "{%f,%f},", g.tri_pt_pos[i][0], g.tri_pt_pos[i][1]);
	//}
	//fseek(f, -1, SEEK_CUR);
	//fprintf(f, "}");
	//fclose(f);

	// clear the intermediate data that was used
	free(in.pointlist);
	free(in.pointmarkerlist);
	free(in.segmentlist);
	free(in.segmentmarkerlist);
}

static vect2f findZero(vect2f pa, vect2f pb, float va, float vb)
{
	return ( pb * va - pa * vb ) / ( va - vb );
}

void marching_squares(vector<vect2f> &contour, vector<double> &bary, int func_idx, float avg_val)
{
#define pb(x,y) bary[(((y)*(g.N+1))+(x))*n + func_idx]
#define makepos(size,x,y) (vect2f(x, y)*(size))
#define sign(x) ((x) < 0 ? -1 : 1)
	contour.clear();
	
	int n = g.poly[0].size();
	float s = 1.0 / g.N;

	for (int x = 0; x < g.N; x++)
	{
		for (int y = 0; y < g.N; y++)
		{
			if (sign(pb(x,y)-avg_val) != sign(pb(x+1,y)-avg_val))
				contour.push_back(findZero(makepos(s,x,y), makepos(s,x+1,y), pb(x,y)-avg_val, pb(x+1,y)-avg_val));
			if (sign(pb(x+1,y)-avg_val) != sign(pb(x+1,y+1)-avg_val))
				contour.push_back(findZero(makepos(s,x+1,y), makepos(s,x+1,y+1), pb(x+1,y)-avg_val, pb(x+1,y+1)-avg_val));
			if (sign(pb(x+1,y+1)-avg_val) != sign(pb(x,y+1)-avg_val))
				contour.push_back(findZero(makepos(s,x+1,y+1), makepos(s,x,y+1), pb(x+1,y+1)-avg_val, pb(x,y+1)-avg_val));
			if (sign(pb(x,y+1)-avg_val) != sign(pb(x,y)-avg_val))
				contour.push_back(findZero(makepos(s,x,y+1), makepos(s,x,y), pb(x,y+1)-avg_val, pb(x,y)-avg_val));
		}
	}
#undef sign
#undef makepos
#undef pb
}

void marching_triangles(vector<vect2f> &contour, vector<vect2i> &contour_edge, int func_idx, float avg_val)
{
#define sign(x) ((x) < 0 ? -1 : 1)
	contour.clear();
	
	int n = g.poly[0].size();
	float s = 1.0 / g.N;


	for (int i = 0; i < g.tri_tris.size(); i++)
	{
		int added = 0;
		for (int j = 0; j < 3; j++)
		{
			int ia = g.tri_tris[i][j];
			int ib = g.tri_tris[i][(j+1)%3];

			vect2d &pa = g.tri_pt_pos[ia];
			vect2d &pb = g.tri_pt_pos[ib];

			double va, vb;
			if (g.deriv == 0 || g.sel_type == 0)
			{
#ifdef QUADRIC_BOUNDARY_DERIV_FIGURE
				va = g.tri_pt_barys[ia*n*3 + func_idx] - avg_val;
				vb = g.tri_pt_barys[ib*n*3 + func_idx] - avg_val;
#else
				va = g.tri_pt_barys[ia*n + func_idx] - avg_val;
				vb = g.tri_pt_barys[ib*n + func_idx] - avg_val;
#endif
			}
			else
			{
				va = g.tri_pt_barys2[ia*n*3 + func_idx*3 + 1] - avg_val;
				vb = g.tri_pt_barys2[ib*n*3 + func_idx*3 + 1] - avg_val;
			}

			if (sign(va) != sign(vb))
			{
				contour.push_back(findZero(pa, pb, va, vb));
				if (ib < ia)
					swap(ia, ib);
				contour_edge.push_back(vect2i(ia, ib));

				added++;
			}
		}
		if ((added != 2 && added != 0) ||
			(added == 2 && (contour[contour.size()-1]-contour[contour.size()-2]).length2() < 1e-12))
		{
			for (int i = 0; i < added; i++)
			{
				contour.pop_back();
				contour_edge.pop_back();
			}
		}
	}

#undef sign
}

vector<Curve> joinLines(vector<vect2f> &contour, vector<vect2i> &contour_edge)
{
	map<vect2i, int> m;
	map<vect2i, vect2f> mp;

	for (int i = 0; i < contour_edge.size(); i++)
	{
		vect2i e = contour_edge[i];
		map<vect2i, int>::iterator it = m.find(e);
		if (it == m.end())
		{
			int &v = m[e];
			v = m.size() - 1;
			mp[e] = contour[i];
		}
	}

	vector< vector<int> > v(m.size());
	vector<bool> used(m.size(), false);
	vector<vect2f> pts(m.size());

	for (map<vect2i, int>::iterator it = m.begin(); it != m.end(); ++it)
		pts[it->second] = mp[it->first];

	for (int i = 0; i < contour_edge.size(); i += 2)
	{
		int a = m[contour_edge[i]];
		int b = m[contour_edge[i+1]];
		v[a].push_back(b);
		v[b].push_back(a);
	}

	// trace all possible contour segments
	vector<Curve> curves;
	for (int i = 0; i < contour_edge.size(); i += 2)
	{
		int a = m[contour_edge[i]];
		int b = m[contour_edge[i+1]];

		if (used[a] && used[b])
			continue;

		vect2i end(a, b);

		Curve c;
		c.pts.push_back(pts[a]);
		c.pts.push_back(pts[b]);
		used[a] = used[b] = true;
		bool going_fwd = true;
		c.closed = true;

		// trace around
		for (int j = 0; j < 2; j++)
		{
			while (true)
			{
				int end_size = v[end[1]].size();
				if (end_size == 1)
					c.closed = false;

				// get next pt
				int next;
				for (int k = 0; k < end_size; k++)
				{
					if (v[end[1]][k] == end[0])
					{
						next = v[end[1]][(k+1) % end_size];
						break;
					}
				}
				
				// check if ran out of stuff to follow
				if (used[next])
					break;

				// go on to next segment
				end[0] = end[1];
				end[1] = next;
				used[next] = true;
				if (going_fwd)
					c.pts.push_back(pts[next]);
				else
					c.pts.push_front(pts[next]);
			}

			swap(a, b);
			end(a, b);
			going_fwd = false;
		}

		curves.push_back(c);
	}

	return curves;
}

void saveSVG()
{
	double scale = 1024;

	char fn[1024];
	sprintf(fn, "output.svg");
	FILE *f = fopen(fn, "wb");

	fprintf(f, "<?xml version=\"1.0\" standalone=\"no\"?> <!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n");
	fprintf(f, "<svg width=\"1024\" height=\"1024\" version=\"1.1\" xmlns=\"http://www.w3.org/2000/svg\">\n");

	// draw contours
	for (int i = 0; i < contour_vals; i++)
	{
		string color;
		if (i < contour_zero) // less than 0
			color = "blue";
		else if (i == contour_zero) // equal to 0
			color = "black";
		else if (i >= contour_zero + contour_steps)
			color = "yellow";
		else // greater than 0
			color = "red";

		for (int j = 0; j < g.curves[i].size(); j++)
		{
			list<vect2f> &ll = g.curves[i][j].pts;

			if (ll.size() > 0)
			{
				fprintf(f, "<path d=\"");
				fprintf(f, "M%f %f\n", ll.front()[0] * scale, scale - ll.front()[1] * scale);
				for (list<vect2f>::iterator k = ll.begin(); k != ll.end(); ++k)
					fprintf(f, "L%f %f\n", (*k)[0] * scale, scale - (*k)[1] * scale);

				fprintf(f, "%c\" style=\"fill:none;stroke:%s;stroke-width:4\"/>\n", g.curves[i][j].closed ? 'Z' : ' ',color.c_str());
			}
		}
	}

	// draw poly
	fprintf(f, "<path d=\"");
	fprintf(f, "M%f %f\n", g.poly[0][0][0] * scale, scale - g.poly[0][0][1] * scale);
	for (int i = 1; i < g.poly[0].size(); i++)
		fprintf(f, "L%f %f\n", g.poly[0][i][0] * scale, scale - g.poly[0][i][1] * scale);
	fprintf(f, "Z\" style=\"fill:none;stroke:green;stroke-width:4\"/>\n");

	// draw selected basis function
	if (g.sel_type == 0)
		fprintf(f, "<circle cx=\"%f\" cy=\"%f\" r=\"8\" style=\"fill:gray;stroke:black;stroke-width:4\"/>", scale*g.poly[0][g.sel_vert][0], scale - scale*g.poly[0][g.sel_vert][1]);
	else
	{
		vect2d p0 = g.poly[0][g.sel_vert];
		vect2d p1 = g.poly[0][(g.sel_vert + 1) % g.poly[0].size()];
		vect2d norm(-p1[1]+p0[1], p1[0]-p0[0]);
		norm.normalize();

		vect2d start = (p0 + p1) * .5;
		vect2d end = start + norm * .05;

		fprintf(f, "<line x1=\"%f\" y1=\"%f\" x2=\"%f\" y2=\"%f\" style=\"fill:gray;stroke:black;stroke-width:4\"/>", scale*start[0], scale - scale*start[1], scale*end[0], scale - scale*end[1]);
	}

	fprintf(f, "</svg>\n");
	fclose(f);
}

void calcContours()
{
	int n = g.poly[0].size();
	g.contours.clear();
	//g.contours.resize(n);
	g.curves.clear();
	//g.curves.resize(n);
	g.contours_edges.clear();
	//g.contours_edges.resize(n);

	//for (int i = 0; i < n; i++)
	{
		g.contours.resize(contour_vals);
		g.curves.resize(contour_vals);
		g.contours_edges.resize(contour_vals);
		for (int v = 0; v < contour_vals; v++)
		{
			float vv = (v - contour_zero)/double(contour_steps);
			//marching_squares(g.contours[i][v], g.barycoords, i, vv);
			marching_triangles(g.contours[v], g.contours_edges[v], g.sel_vert, vv);

			g.curves[v] = joinLines(g.contours[v], g.contours_edges[v]);
		}
	}

	//saveSVG();
}

void calcDeformation()
{
	for (int x = 0; x <= g.N; x++)
	{
		for (int y = 0; y <= g.N; y++)
		{
			int n = g.poly[0].size();
			vect2d &q = g.deformation[(y*(g.N+1))+x];
			q(0, 0);

			double *bary = &g.barycoords[((y*(g.N+1))+x)*n];
			for (int i = 0; i < n; i++)
				q += g.poly[1][i]*bary[i];

			double *bary2 = &g.barycoords2[((y*(g.N+1))+x)*g.norms[0].size()];
			for (int i = 0; i < n; i++)
				q += g.norms[1][i]*bary2[i];
		}
	}
}

void calc_surf()
{
#if 0
	g.surf.resize(g.N+2, g.N+2);

	for (int x = 0; x <= g.N; x++)
	{
		for (int y = 0; y <= g.N; y++)
		{
			int n = g.poly[0].size();
			double *bary = &g.barycoords[((y*(g.N+1))+x)*n];
			double h = 0;
			for (int i = 0; i < n; i++)
				h += g.height_vals[i]*bary[i];

			double *bary2 = &g.barycoords2[((y*(g.N+1))+x)*g.norms[0].size()];
			for (int i = 0; i < n; i++)
				h += g.deriv_vals[i]*bary2[i];

			double xx = double(x)/g.N;
			double yy = double(y)/g.N;
			g.surf(x,y)(xx, yy, h);
		}
	}

	// set edge borders
	for (int i = 0; i <= g.N+1; i++)
		g.surf(i,g.N+1)(double(i)/g.N, double(g.N+1)/g.N, g.surf(i,g.N)[2]);
	
	for (int i = 0; i <= g.N+1; i++)
		g.surf(g.N+1,i)(double(g.N+1)/g.N, double(i)/g.N, g.surf(g.N,i)[2]);
#endif
}

double getValue(int x, int y)
{
	int n = g.poly[0].size();
	double *bary = &g.barycoords[((y*(g.N+1))+x)*n];

	return bary[g.sel_vert];
}

void setColor(int x, int y)
{
	double v = getValue(x,y);

	if (v < 0)
		glColor3f(-v,-v,1);
	else if (v > 1)
		glColor3f(1,v-1,v-1);
	else
	{
		glColor3f(v,v,v);
	}
}

void save_cylinder(FILE *f, vect3f p0, vect3f p1, float radius, int sides, int &verts_written)
{
	// frame with z along cylinder
	vect3f z = ~(p1 - p0);
	vect3f x = ~(z % vect3f(1,1,1));
	vect3f y = z % x;

	// output a cylinder
	for (int i = 0; i < sides; i++)
	{
		float theta = (2 * pi * i) / sides;
		float theta_p = (2 * pi * ((i + 1) % sides)) / sides;

		vect3f v0 = (x*cos(theta) + y*sin(theta)) * radius;
		vect3f v1 = (x*cos(theta_p) + y*sin(theta_p)) * radius;

		vect3f a0 = p0 + v0;
		vect3f a1 = p0 + v1;
		vect3f a2 = p1 + v1;
		vect3f a3 = p1 + v0;
		fprintf(f, "v %f %f %f\n", a0[0], a0[1], a0[2]);
		fprintf(f, "v %f %f %f\n", a1[0], a1[1], a1[2]);
		fprintf(f, "v %f %f %f\n", a2[0], a2[1], a2[2]);
		fprintf(f, "v %f %f %f\n", a3[0], a3[1], a3[2]);
		fprintf(f, "f %d %d %d\n", verts_written + 1, verts_written + 2, verts_written + 4);
		fprintf(f, "f %d %d %d\n", verts_written + 2, verts_written + 3, verts_written + 4);

		verts_written += 4;
	}
}

void save_sphere(FILE *f, vect3f center, float radius, int sides, int &verts_written)
{
	vect3f x(1,0,0);
	vect3f y(0,1,0);
	vect3f z(0,0,1);

	for (int j = 0; j < sides; j++)
	{
		float phi = (pi * j) / sides;
		float phi_p = (pi * (j+1)) / sides;

		for (int i = 0; i < sides; i++)
		{
			float theta = (2 * pi * i) / sides;
			float theta_p = (2 * pi * ((i + 1) % sides)) / sides;

			vect3f v0 = x*cos(theta) + y*sin(theta);
			vect3f v1 = x*cos(theta_p) + y*sin(theta_p);

			vect3f a0 = center + (z * -cos(phi) + v0 * sin(phi)) * radius;
			vect3f a1 = center + (z * -cos(phi) + v1 * sin(phi)) * radius;
			vect3f a2 = center + (z * -cos(phi_p) + v1 * sin(phi_p)) * radius;
			vect3f a3 = center + (z * -cos(phi_p) + v0 * sin(phi_p)) * radius;
			fprintf(f, "v %f %f %f\n", a0[0], a0[1], a0[2]);
			fprintf(f, "v %f %f %f\n", a1[0], a1[1], a1[2]);
			fprintf(f, "v %f %f %f\n", a2[0], a2[1], a2[2]);
			fprintf(f, "v %f %f %f\n", a3[0], a3[1], a3[2]);
			fprintf(f, "f %d %d %d\n", verts_written + 1, verts_written + 2, verts_written + 4);
			fprintf(f, "f %d %d %d\n", verts_written + 2, verts_written + 3, verts_written + 4);

			verts_written += 4;
		}
	}
}

void save_boundary(vector<vect3d> &pts, vector<vect3d> &norms)
{
	const float radius = .003;
	const int sides = 16;
	int verts_written = 0;

	FILE *f = fopen("output_boundary.obj", "w");

	// output cylinders
	for (int k = 0; k < g.tri_edges.size(); k++)
	{
		if (g.tri_pt_border[g.tri_edges[k][0]] && g.tri_pt_border[g.tri_edges[k][1]])
		{
			// end points
			vect3f p0 = pts[g.tri_edges[k][0]];
			vect3f p1 = pts[g.tri_edges[k][1]];

			//p0[2] = g.tri_pt_barys[g.tri_edges[k][0]*3];
			//p1[2] = g.tri_pt_barys[g.tri_edges[k][1]*3];

			save_cylinder(f, p0, p1, radius, sides, verts_written);
		}
	}

	// write spheres at vertices
	for (int k = 0; k < pts.size(); k++)
	{
		if (!g.tri_pt_border[k])
			continue;

		save_sphere(f, pts[k], radius, sides, verts_written);
	}

	fclose(f);
}

void save_axes(vector<vect3d> &pts, vector<vect3d> &norms)
{
	const float radius = .005;
	const int sides = 16;
	int verts_written = 0;

	FILE *f = fopen("output_axes.obj", "w");

	// output cylinders
	save_cylinder(f, vect3f(-0.1,-0.1,0), vect3f(1,-0.1,0), radius, sides, verts_written);
	save_cylinder(f, vect3f(-0.1,-0.1,0), vect3f(-0.1,1,0), radius, sides, verts_written);
	save_cylinder(f, vect3f(-0.1,-0.1,0), vect3f(-0.1,-0.1,.4), radius, sides, verts_written);

	// write spheres at vertices
	save_sphere(f, vect3f(-0.1,-0.1,0), radius, sides, verts_written);

	fclose(f);
}

void save_graph(vector<vect3d> &pts, vector<vect3d> &norms)
{
	// save function graph
	FILE *out = fopen("output.obj", "w");

	for (int i = 0; i < pts.size(); i++)
		fprintf(out, "v %f %f %f\n", pts[i][0], pts[i][1], pts[i][2]);

	for (int i = 0; i < norms.size(); i++)
		fprintf(out, "vn %f %f %f\n", norms[i][0], norms[i][1], norms[i][2]);

	for (int i = 0; i < g.tri_tris.size(); i++)
		fprintf(out, "f %d %d %d\n", g.tri_tris[i][0] + 1, g.tri_tris[i][1] + 1, g.tri_tris[i][2] + 1);

	fclose(out);
}

void draw()
{
	//// calculate the deformation
	//calcDeformation();

	// draw
	glClearColor(1,1,1,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//** input poly
	glViewport(0, 0, g.scr_width/2, g.scr_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 1, 0, 1, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// draw triangles
	glLineWidth(1);
	glColor3f(.5,0,.7);
	glBegin(GL_LINES);
	for (int i = 0; i < g.tri_edges.size(); i++)
	{
		glVertex2dv(g.tri_pt_pos[g.tri_edges[i][0]].v);
		glVertex2dv(g.tri_pt_pos[g.tri_edges[i][1]].v);
	}
	glEnd();

	// image
	if (g.draw_surface)
	{
		if (g.draw_contours)
		{
			//if (g.contours.size() == g.poly[0].size())
			{
				glLineWidth(3);
				glPointSize(3);
				for (int i = 0; i < contour_vals; i++)
				{
					if (i < contour_zero) // less than 0
						glColor3f(0,0,1);
					else if (i == contour_zero) // equal to 0
						glColor3f(0,0,0);
					else if (i >= contour_zero + contour_steps)
						glColor3f(1,1,0);
					else // greater than 0
						glColor3f(1,0,0);

#if 0
					glBegin(GL_LINES);
					for (int j = 0; j < g.contours[g.sel_vert][i].size(); j++)
						glVertex2fv(g.contours[g.sel_vert][i][j].v);
					glEnd();
#else
					for (int j = 0; j < g.curves[i].size(); j++)
					{
						if (g.curves[i][j].closed)
							glBegin(GL_LINE_LOOP);
						else
							glBegin(GL_LINE_STRIP);

						list<vect2f> &ll = g.curves[i][j].pts;
						for (list<vect2f>::iterator k = ll.begin(); k != ll.end(); ++k)
							glVertex2fv(k->v);

						glEnd();
					}
#endif

#if 0
					glBegin(GL_POINTS);
					for (int j = 0; j < g.contours[g.sel_vert][i].size(); j++)
						glVertex2fv(g.contours[g.sel_vert][i][j].v);
					glEnd();
#endif
				}
			}
		}
		else
		{
			for (int x = 0; x < g.N; x++)
			{
				float xl = (float)x/g.N;
				float xh = (float)(x+1)/g.N;

				glBegin(GL_TRIANGLE_STRIP);
				for (int y = 0; y <= g.N; y++)
				{
					float yy = (float)y/g.N;

					setColor(x, y);
					glVertex2f(xl, yy);

					setColor(x+1, y);
					glVertex2f(xh, yy);
				}
				glEnd();
			}
		}
	}
	else if (g.draw_image)
	{
		glEnable(GL_TEXTURE_2D);
		glColor3f(1,1,1);
		g.img->bind();
		for (int x = 0; x < g.N; x++)
		{
			float xl = (float)x/g.N;
			float xh = (float)(x+1)/g.N;

			glBegin(GL_TRIANGLE_STRIP);
			for (int y = 0; y <= g.N; y++)
			{
				float yy = (float)y/g.N;

				glTexCoord2f(xl, yy);
				glVertex2f  (xl, yy);

				glTexCoord2f(xh, yy);
				glVertex2f  (xh, yy);
			}
			glEnd();
		}
		glDisable(GL_TEXTURE_2D);
	}

	// grid
	if (g.draw_grid)
	{
		glColor3f(0,0,.3);
		
		for (int x = 0; x < g.N; x++)
		{
			glBegin(GL_LINES);
			for (int y = 0; y < g.N; y++)
			{
				vect2d b(double(x)/g.N, double(y)/g.N);
				vect2d t(double(x+1)/g.N, double(y+1)/g.N);

				glVertex2d(b[0], b[1]);
				glVertex2d(b[0], t[1]);

				glVertex2d(b[0], t[1]);
				glVertex2d(t[0], t[1]);
			}
			glEnd();
		}
	}

	// poly
	glLineWidth(2); // lines
	glColor3f(.2,.8,0);
	glBegin(GL_LINE_LOOP);
	for (int i = 0; i < g.poly[0].size(); i++)
		glVertex2dv(g.poly[0][i].v);
	glEnd();

	glPointSize(3); // vertex control pts
	glColor3f(1,0,0);
	glBegin(GL_POINTS);
	for (int i = 0; i < g.poly[0].size(); i++)
		glVertex2dv(g.poly[0][i].v);
	glEnd();

	if (g.deriv)
	{
		glLineWidth(1); // normal lines
		glColor3f(.5,.5,0);
		glBegin(GL_LINES);
		for (int i = 0; i < g.poly[0].size(); i++)
		{
			int ip = (i+1)%g.poly[0].size();
			vect2d c = (g.poly[0][ip] + g.poly[0][i]) * .5;

			glVertex2dv(c.v);
			glVertex2dv((c+g.norms[0][i]*0.05).v);
		}
		glEnd();
		
		glLineWidth(1); // normal control pts
		glColor3f(0,0,1);
		glBegin(GL_POINTS);
		for (int i = 0; i < g.poly[0].size(); i++)
		{
			int ip = (i+1)%g.poly[0].size();
			vect2d c = (g.poly[0][ip] + g.poly[0][i]) * .5;

			glVertex2dv((c+g.norms[0][i]*0.05).v);
		}
		glEnd();
	}

	//** RIGHT WINDOW
	if (g.draw_surface)
	{
		glViewport(g.scr_width/2, 0, g.scr_width/2, g.scr_height);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(60, 1, .1, 10);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		gluLookAt(2,0,.5,
				  .50,.50,0,
				  0,0,1);

		glTranslatef(.5,.5,.5);
		glRotatef(SDL_GetTicks()/200.0, 0, 0, 1);
		glTranslatef(-.5,-.5,-.5);

		// calc the surface
		calc_surf();

		// surface
		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_TEXTURE_2D);

		glColor3f(1,0,0);
		if (true)
		{
			int n = g.poly[0].size();

			// calc normals
			vector<vect3d> norms, pts;
			norms.resize(g.tri_pt_pos.size());
			pts.resize(g.tri_pt_pos.size());


			if (g.use_basis_values)
			{
				for (int i = 0; i < pts.size(); i++)
				{
#ifdef QUADRIC_BOUNDARY_DERIV_FIGURE
					double val = 0;

					/*for (int k = 0; k < n*3; k++)
						val += g.height_vals[k] * g.tri_pt_barys[i*n*3 + k];

					if (g.deriv)
					{
						for (int k = 0; k < n*3; k++)
							val += g.deriv_vals[k] * g.tri_pt_barys2[i*n*3 + k];
					}*/

					val += g.tri_pt_barys[i*n*3 + 0];
					val += g.tri_pt_barys[i*n*3 + n*3-1];

					pts[i](g.tri_pt_pos[i][0], g.tri_pt_pos[i][1], g.height_scale * val);
#else
					// hard coded values for making a figure
					pts[i](g.tri_pt_pos[i][0], g.tri_pt_pos[i][1], 0);

					/*pts[i][2] += g.tri_pt_barys[i*n + 0] * .25;
					pts[i][2] += g.tri_pt_barys[i*n + 1] * .4;
					pts[i][2] += g.tri_pt_barys[i*n + 2] * .3;
					pts[i][2] += g.tri_pt_barys[i*n + 3] * .5;
					pts[i][2] += g.tri_pt_barys[i*n + 4] * .4;*/

					pts[i][2] += g.tri_pt_barys[i*n + 0];
#endif
					norms[i] = 0;
				}
			}
			else if (g.deriv == 0 || g.sel_type == 0)
			{
				for (int i = 0; i < pts.size(); i++)
				{
#ifdef QUADRIC_BOUNDARY_DERIV_FIGURE
					pts[i](g.tri_pt_pos[i][0], g.tri_pt_pos[i][1], g.height_scale *  g.tri_pt_barys[i*n*3 + g.sel_vert]);
#else
					pts[i](g.tri_pt_pos[i][0], g.tri_pt_pos[i][1], g.height_scale * g.tri_pt_barys[i*n + g.sel_vert]);
#endif
					norms[i] = 0;
				}
			}
			else
			{
				for (int i = 0; i < pts.size(); i++)
				{
					pts[i](g.tri_pt_pos[i][0], g.tri_pt_pos[i][1], g.height_scale * g.tri_pt_barys2[i*n*3 + g.sel_vert*3 + 1]);
					norms[i] = 0;
				}
			}

			for (int i = 0; i < g.tri_tris.size(); i++)
			{
				vect3d N = (pts[g.tri_tris[i][1]] - pts[g.tri_tris[i][0]]) % (pts[g.tri_tris[i][2]] - pts[g.tri_tris[i][0]]);
				for (int j = 0; j < 3; j++)
					norms[g.tri_tris[i][j]] += N;
			}

			for (int i = 0; i < pts.size(); i++)
				norms[i] /= norms[i].length();

			// draw tris
			glBegin(GL_TRIANGLES);
			for (int i = 0; i < g.tri_tris.size(); i++)
			{
				for (int j = 0; j < 3; j++)
				{
					int k = g.tri_tris[i][j];

					glNormal3dv(norms[k].v);
					glTexCoord2dv(g.tri_pt_pos[k].v);
					glVertex3dv(pts[k].v);
				}
			}
			glEnd();

#ifdef SAVE_OBJ_AND_EXIT // save obj file to disk and exit
			static bool did = false;
			if (!did)
			{
				did = true;
				save_boundary(pts, norms);
				save_graph(pts, norms);
				save_axes(pts, norms);
				exit(1);
			}
#endif
		}
		else
		{
			for (int x = 0; x < g.N-1; x++)
			{
				double xl = (double)x/g.N;
				double xh = (double)(x+1)/g.N;

				glBegin(GL_TRIANGLE_STRIP);
				for (int y = 0; y <= g.N-1; y++)
				{
					glNormal3dv(( ~((g.surf(x+1,y)-g.surf(x,y))%(g.surf(x,y+1)-g.surf(x,y))) ).v);
					glTexCoord2dv(g.surf(x,y).v);
					glVertex3dv(g.surf(x,y).v);

					glNormal3dv(( ~((g.surf(x+2,y)-g.surf(x+1,y))%(g.surf(x+1,y+1)-g.surf(x+1,y))) ).v);
					glTexCoord2dv(g.surf(x+1,y).v);
					glVertex3dv(g.surf(x+1,y).v);
				}
				glEnd();
			}
		}
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);

		// poly
		glLineWidth(4); // lines
		glColor3f(1,1,0);
		glBegin(GL_LINE_LOOP);
		for (int i = 0; i < g.poly[1].size(); i++)
		{
			if (g.use_basis_values)
			{
#ifdef QUADRIC_BOUNDARY_DERIV_FIGURE
				//glVertex3d(g.poly[0][i][0], g.poly[0][i][1], g.height_scale * g.height_vals[i*3]);
#endif
			}
			else
			{
				glVertex3d(g.poly[0][i][0], g.poly[0][i][1], i == g.sel_vert && (g.deriv == 0 || g.sel_type == 0) ? g.height_scale : 0);
			}
		}
		glEnd();

		glDisable(GL_DEPTH_TEST);
	}
	else
	{
		glViewport(g.scr_width/2, 0, g.scr_width/2, g.scr_height);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0, 1, 0, 1, -1, 1);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		// image
		if (g.draw_image)
		{
			glEnable(GL_TEXTURE_2D);
			glColor3f(1,1,1);
			g.img->bind();
			for (int x = 0; x < g.N; x++)
			{
				double xl = (double)x/g.N;
				double xh = (double)(x+1)/g.N;

				glBegin(GL_TRIANGLE_STRIP);
				for (int y = 0; y <= g.N; y++)
				{
					double yy = (double)y/g.N;

					glTexCoord2d(xl, yy);
					glVertex2dv (g.deformation[(y)*(g.N+1)+x].v);

					glTexCoord2d(xh, yy);
					glVertex2dv (g.deformation[(y)*(g.N+1)+x+1].v);
				}
				glEnd();
			}
			glDisable(GL_TEXTURE_2D);
		}

		// grid
		if (g.draw_grid)
		{
			glColor3f(0,0,.3);
			g.img->bind();
			for (int x = 0; x < g.N; x++)
			{
				glBegin(GL_LINES);
				for (int y = 0; y < g.N; y++)
				{
					glVertex2dv(g.deformation[(y)*(g.N+1)+x].v);
					glVertex2dv(g.deformation[(y)*(g.N+1)+x+1].v);

					glVertex2dv(g.deformation[(y+1)*(g.N+1)+x].v);
					glVertex2dv(g.deformation[(y)*(g.N+1)+x].v);
				}
				glEnd();
			}
		}

		// poly
		glLineWidth(1); // lines
		glColor3f(1,1,0);
		glBegin(GL_LINE_LOOP);
		for (int i = 0; i < g.poly[1].size(); i++)
			glVertex2dv(g.poly[1][i].v);
		glEnd();

		glPointSize(3); // vertex control pts
		glColor3f(1,0,0);
		glBegin(GL_POINTS);
		for (int i = 0; i < g.poly[1].size(); i++)
			glVertex2dv(g.poly[1][i].v);
		glEnd();

		if (g.deriv)
		{
			glLineWidth(1); // normal lines
			glColor3f(.5,.5,0);
			glBegin(GL_LINES);
			for (int i = 0; i < g.poly[1].size(); i++)
			{
				int ip = (i+1)%g.poly[1].size();
				vect2d c = (g.poly[1][ip] + g.poly[1][i]) * .5;

				glVertex2dv(c.v);
				glVertex2dv((c+g.norms[1][i]*0.05).v);
			}
			glEnd();

			glLineWidth(1); // normal control pts
			glColor3f(0,0,1);
			glBegin(GL_POINTS);
			for (int i = 0; i < g.poly[1].size(); i++)
			{
				int ip = (i+1)%g.poly[1].size();
				vect2d c = (g.poly[1][ip] + g.poly[1][i]) * .5;

				glVertex2dv((c+g.norms[1][i]*0.05).v);
			}
			glEnd();
		}
	}

	// middle line
	glViewport(0, 0, g.scr_width, g.scr_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 2, 0, 1, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glLineWidth(2);
	glColor3f(0,0,1);
	glBegin(GL_LINES);
	glVertex2f(1, 0);
	glVertex2f(1, 1);
	glEnd();
}

void logic()
{
	// keep points in [0,1] box
	for (int n = 0; n < 2; n++)
		for (int i = 0; i < g.poly[n].size(); i++)
			for (int j = 0; j < 2; j++)
				g.poly[n][i][j] = min(1, max(0, g.poly[n][i][j]));
}

void ensureCorrectness()
{
	// add missing normals to the control poly
	for (int i = g.norms[0].size(); i < g.poly[0].size(); i++)
	{
		int ip = (i+1)%g.poly[0].size();
		g.norms[0].push_back(~perp(g.poly[0][ip] - g.poly[0][i]));
	}

	// fix normals of input poly
	for (int i = 0; i < g.poly[0].size(); i++)
	{
		int ip = (i+1)%g.poly[0].size();
		g.norms[0][i] = ~perp(g.poly[0][ip] - g.poly[0][i]);
	}

	// add any missing normals to the deformed poly
	for (int i = g.norms[1].size(); i < g.poly[1].size(); i++)
	{
		int ip = (i+1)%g.poly[1].size();
		g.norms[1].push_back(~perp(g.poly[1][ip] - g.poly[1][i]));
	}

	//// create height vals
	//for (int i = g.height_vals.size(); i < g.poly[0].size(); i++)
	//	g.height_vals.push_back(0);
	//
	//// create deriv vals
	//for (int i = g.deriv_vals.size(); i < g.poly[0].size(); i++)
	//	g.deriv_vals.push_back(0);
}

void input()
{
	ensureCorrectness();

	bool calc_bary = false;
	bool calc_contours = false;

	SDL_Event event;
	while (SDL_PollEvent(&event))
	{
		if (event.type == SDL_KEYDOWN)
		{
			bool k_ctrl = event.key.keysym.mod & (KMOD_LCTRL | KMOD_RCTRL);
			bool k_alt = event.key.keysym.mod & (KMOD_LALT | KMOD_RALT);
			bool k_shift = event.key.keysym.mod & (KMOD_LSHIFT | KMOD_RSHIFT);

			if (event.key.keysym.sym == SDLK_ESCAPE)
				g.running = false;
			else if (event.key.keysym.sym == SDLK_SPACE)
				calc_bary = true;
			else if (event.key.keysym.sym == SDLK_f)
			{
				g.use_basis_values = !g.use_basis_values;
			}
			else if (event.key.keysym.sym == SDLK_d)
			{
				g.deriv = (g.deriv + 1) % 2;
				printf("deriv = %d\n", g.deriv);
				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_a)
			{
				g.alpha++;
				if (g.alpha == 3)
					g.alpha = 1;
				printf("alpha = %d\n", g.alpha);

				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_1)
			{
				g.method = Method::points_linear;
				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_2)
			{
				g.method = Method::wachspress;
				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_3)
			{
				g.method = Method::lines_linear;
				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_4)
			{
				g.method = Method::lines_quad;
				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_5)
			{
				g.method = Method::mean_value;
				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_6)
			{
				g.method = Method::harmonic;
				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_7)
			{
				g.method = Method::maximum_entropy;
				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_0)
			{
				g.method = Method::from_file;
				calc_bary = true;
			}
			else if (event.key.keysym.sym == SDLK_UP)
			{
				/*if (g.sel_poly == 0 && g.sel_type == 0)
					g.height_vals[g.sel_vert] += .025;
				else if (g.sel_poly == 0 && g.sel_type == 1)
					g.deriv_vals[g.sel_vert] += .025;*/
				
				if (g.sel_poly == 0 && g.sel_type == 0)
					g.height_scale += .025;
				else if (g.sel_poly == 0 && g.sel_type == 1)
					g.deriv_vals[g.sel_vert] += .025;
				
			}
			else if (event.key.keysym.sym == SDLK_DOWN)
			{
				/*if (g.sel_poly == 0 && g.sel_type == 0)
					g.height_vals[g.sel_vert] -= .025;
				else if (g.sel_poly == 0 && g.sel_type == 1)
					g.deriv_vals[g.sel_vert] -= .025;*/
				
				if (g.sel_poly == 0 && g.sel_type == 0)
					g.height_scale -= .025;
				else if (g.sel_poly == 0 && g.sel_type == 1)
					g.deriv_vals[g.sel_vert] -= .025;
			}
			else if (event.key.keysym.sym == SDLK_RETURN)
			{
				g.draw_surface = !g.draw_surface;
			}
		}
		else if (event.type == SDL_QUIT)
		{
			g.running = false;
		}
		else if (event.type == SDL_MOUSEMOTION)
		{
			if (event.motion.state == 4)
			{
				vect2d v(2 * (double)event.motion.xrel / g.scr_width, -(double)event.motion.yrel / g.scr_height);
				if (g.sel_type == 0)
				{
					g.poly[g.sel_poly][g.sel_vert] += v;
				}
				else
				{
					g.norms[g.sel_poly][g.sel_vert] += v/.05;
				}
			}
		}
		else if (event.type == SDL_MOUSEBUTTONDOWN)
		{
			if (event.button.button == 1)
			{
				vect2d p(2 * (double)event.button.x / g.scr_width, 1 - (double)event.button.y / g.scr_height);
				g.poly[0].push_back(p);
				g.poly[1].push_back(p);
				ensureCorrectness();
				calc_bary = true;
			}
			else if (event.button.button == 3)
			{
				calc_contours = true;

				vect2d p(2 * (double)event.button.x / g.scr_width, 1 - (double)event.button.y / g.scr_height);
				g.sel_poly = p[0];
				p[0] -= g.sel_poly;

				// select control point
				g.sel_type = 0;
				g.sel_vert = 0;
				double dist_min = 1e30;
				for (int i = 0; i < g.poly[g.sel_poly].size(); i++) // vertex
				{
					double dist = !(g.poly[g.sel_poly][i] - p);
					if (dist < dist_min)
					{
						dist_min = dist;
						g.sel_vert = i;
						g.sel_type = 0;
					}
				}
				for (int i = 0; i < g.poly[g.sel_poly].size(); i++) // normal
				{
					int ip = (i+1)%g.poly[g.sel_poly].size();
					double dist = !((g.poly[g.sel_poly][ip]+g.poly[g.sel_poly][i])*.5 + g.norms[1][i]*.05 - p);
					if (dist < dist_min)
					{
						dist_min = dist;
						g.sel_vert = i;
						g.sel_type = 1;
					}
				}
			}
		}
	}
	
	if (calc_bary)
	{
		double t_start = SDL_GetTicks();
		calcTriangulation();
		calcBaryCoords();
		calc_contours = true;
		printf("coordinates in %f msec\n", SDL_GetTicks() - t_start);
	}

	if (calc_contours)
		calcContours();
}

#ifdef QUADRIC_BOUNDARY_DERIV_FIGURE
void fix_deriv_vals()
{
	// cross boundary derivatives are constrained by the function values on the boundary
	vector<vect2d> &verts = g.poly[0];
	int n = verts.size();

	for (int i = 0; i < n; i++)
	{
		// set the derivative of the first control point
		vect3d v0(
			verts[(i+1)%n][0] - verts[i][0], 
			verts[(i+1)%n][1] - verts[i][1], 
			-2*g.height_vals[(i*3 + 0) % (3*n)] + 2*g.height_vals[(i*3 + 1) % (3*n)]);
		vect3d v1(
			verts[(i-1+n)%n][0] - verts[i][0], 
			verts[(i-1+n)%n][1] - verts[i][1], 
			-2*g.height_vals[(i*3 - 1 + 3*n) % (3*n)] + 2*g.height_vals[(i*3 - 2 + 3*n) % (3*n)]);

		vect3d norm = v0 % v1;
		vect2d grad(-norm[0] / norm[2], -norm[1] / norm[2]);

		{
		vect2d perp(-v0[1], v0[0]);
		double val = perp * grad;
		g.deriv_vals[i*3 + 0] = val;
		}
		
		{
		vect2d perp(v1[1], -v1[0]);
		double val = perp * grad;
		g.deriv_vals[(i*3 - 1 + 3*n) % (3*n)] = val;
		}
	}
}
#endif

void init()
{
	// init SDL
	SDL_Init(SDL_INIT_VIDEO);
	g.screen = SDL_SetVideoMode(g.scr_width, g.scr_height, 32, SDL_HWSURFACE | SDL_OPENGL | SDL_HWACCEL | (g.fullscreen ? SDL_FULLSCREEN : 0));
	SDL_EnableUNICODE(1);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

	// some gl presets
	initOpenglFuncs();

#if 0
	// mathematica friendly (spiral)
	const int num_pts = 10;
	double pts[num_pts][2] = {{0., 0.}, {0.25, 0.}, {0.25, 0.5}, {0.75, 0.5}, {0.75, 0.25}, {0.5, 0.25}, {0.5, 0.}, {1., 0.}, {1., 0.75}, {0., 0.75}};
	for (int i = 0; i < num_pts; i++)
		g.poly[0].push_back(vect2d(pts[i][0],pts[i][1]));
#elif 1
	// 5 sided concave test
	const int num_pts = 5;
	double pts[num_pts][2] = {{0.1, 0.2}, {.5, 0}, {1, .7}, {.5, .5}, {0, 1}};
	for (int i = 0; i < num_pts; i++)
		g.poly[0].push_back(vect2d(pts[i][0],pts[i][1]));
#elif 0
	// regular n-sided polygons
	int sides = 3;
	for (int i = 0; i < sides; i++)
	{
		vect2d pos(cos(float(i)/sides * 2*pi)*.4 + .5, sin(float(i)/sides * 2*pi)*.4 + .5);
		g.poly[0].push_back(pos);
	}
#elif 0
	// square
	g.poly[0].push_back(vect2d(0,0));
	g.poly[0].push_back(vect2d(1,0));
	g.poly[0].push_back(vect2d(1,1));
	g.poly[0].push_back(vect2d(0,1));


	g.height_vals.push_back(1);
	g.height_vals.push_back(3);
	g.height_vals.push_back(2);
	
	g.height_vals.push_back(2);
	g.height_vals.push_back(1);
	g.height_vals.push_back(1);
	
	g.height_vals.push_back(1);
	g.height_vals.push_back(3);
	g.height_vals.push_back(2);
	
	g.height_vals.push_back(2);
	g.height_vals.push_back(0);
	g.height_vals.push_back(1);
	

	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(20);
	g.deriv_vals.push_back(666);
	
	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(-20);
	g.deriv_vals.push_back(666);
	
	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(20);
	g.deriv_vals.push_back(666);
	
	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(-20);
	g.deriv_vals.push_back(666);
	
	fix_deriv_vals();
#elif 0
	// triangle
	g.poly[0].push_back(vect2d(0,0));
	g.poly[0].push_back(vect2d(1,0));
	g.poly[0].push_back(vect2d(0,1));


	g.height_vals.push_back(1);
	g.height_vals.push_back(1);
	g.height_vals.push_back(1);
	
	g.height_vals.push_back(1);
	g.height_vals.push_back(1);
	g.height_vals.push_back(1);
	
	g.height_vals.push_back(1);
	g.height_vals.push_back(3);
	g.height_vals.push_back(1);
	

	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(0);
	g.deriv_vals.push_back(666);
	
	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(0);
	g.deriv_vals.push_back(666);
	
	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(0);
	g.deriv_vals.push_back(666);
	
	fix_deriv_vals();
#else
	// hexagon
	const int num_pts = 6;
	double pts[num_pts][2] = {{0.5, 0.866025}, {-0.5, 0.866025}, {-1., 0.}, {-0.5, -0.866025}, {0.5, -0.866025}, {1., 0.}};
	for (int i = 0; i < num_pts; i++)
		g.poly[0].push_back(vect2d(pts[i][0],pts[i][1]));


	g.height_vals.push_back(0);
	g.height_vals.push_back(0);
	g.height_vals.push_back(0);
	
	g.height_vals.push_back(0);
	g.height_vals.push_back(0);
	g.height_vals.push_back(0);
	
	g.height_vals.push_back(0);
	g.height_vals.push_back(0);
	g.height_vals.push_back(0);
	
	g.height_vals.push_back(0);
	g.height_vals.push_back(0);
	g.height_vals.push_back(0);

	g.height_vals.push_back(0);
	g.height_vals.push_back(0);
	g.height_vals.push_back(0);

	g.height_vals.push_back(0);
	g.height_vals.push_back(0);
	g.height_vals.push_back(0);

	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(0);
	g.deriv_vals.push_back(666);
	
	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(0);
	g.deriv_vals.push_back(666);
	
	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(0);
	g.deriv_vals.push_back(666);
	
	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(0);
	g.deriv_vals.push_back(666);

	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(0);
	g.deriv_vals.push_back(666);

	g.deriv_vals.push_back(666);
	g.deriv_vals.push_back(0);
	g.deriv_vals.push_back(666);
	
	fix_deriv_vals();
#endif
	g.poly[1] = g.poly[0];

	ensureCorrectness();

	calcTriangulation();
	calcBaryCoords();
	calcContours();

	// load image
	g.img = new Image("checker_small.png");
	//g.img = new Image("hurricane.png");
	//g.img = new Image("gecko.png");
}

void run()
{
	logic();
	g.time = SDL_GetTicks();

	while (g.running)
	{
		double sdl_time = SDL_GetTicks();
		while (g.time < sdl_time && g.running)
		{
			g.time += 10;

			input();
			logic();
		}

		draw();

		SDL_GL_SwapBuffers();

		// simple framerate calculation
		double frame_time = SDL_GetTicks() - sdl_time;
		g.frame_time = g.frame_time * .9 + frame_time * .1;
	}
}

void halt()
{
	SDL_Quit();
}

int main(int argc, char **argv)
{
	init();
	run();
	halt();

	return 0;
}
